> module Feed (Feed(..), Item(..), Status(..), Restart(..), newFeed, monitorFeed) where

> import Control.Monad
> import Control.Concurrent
> import Control.Concurrent.STM
> import qualified Control.Exception as E
> import Data.Hash.MD5
> import Data.Typeable
> import List
> import Maybe
> import Network.Browser
> import Network.HTTP
> import Network.HTTP.Headers
> import Network.URI
> import Text.XML.HaXml.Parse
> import Text.XML.HaXml.Pretty
> import Text.XML.HaXml.Types

> import Util

Our main datatype, what we're primarily working with. Since a feed will need to
be periodically refreshed (based on the 'interval'), we'll create a new thread
for each feed, and it can sleep until it's time to reload.

We'll keep each feed in a TVar, so that the thread fetching the feed and the UI
thread can both get at it. After the feed is created, the UI should not modify
it, other than to delete it.

> data Feed = Feed {
>                   uri :: URI,                 -- where we're connecting to
>                   interval :: Int,            -- how often to connect to them
>                   thread :: Maybe ThreadId,
>                   modified :: Maybe String,   -- last known change time
>                   status :: Status,           -- status from last connection
>                   feedtitle :: Maybe String,
>                   feedlink :: Maybe String,
>                   feeddesc :: Maybe String,
>                   items :: [Item]
>             } deriving Show

When adding a new feed, try to avoid adding a feed we already know about. Also
used for showFeed, to see if we're showing the selected feed.

> instance Eq Feed where
>   a == b = uri a == uri b

The current status of the feed, based on the most recent attempt to pull the
feed.

> data Status = Updating | NoError |
>               NetError | HdrError | XmlError | RssError
>             deriving (Eq, Show, Read)

Items are constructed from the XML and are never modified.

> data Item = Item {
>                   guid :: String,
>                   itemtitle :: Maybe String,
>                   itemlink :: Maybe String,
>                   itemdesc :: Maybe String,
>                   author :: Maybe String
>             } deriving Show

For showItem, to see if we're shoing the selected item.

> instance Eq Item where
>   a == b = guid a == guid b

There's a way for the user to force a feed to be reloaded, and in order to do
that we need to wake up the feed's thread, which means throwing a message to
it.

> data Restart = Restart
> instance Typeable Restart where
>   typeOf _ = mkTyConApp (mkTyCon "Restart") []

Now that we have all of our data declarations out of the way, it's time to get
on to the actual functions in this functional program.

First off, a helper function to make it easy for the UI to create a new feed.

> newFeed :: String -> Int -> Maybe Feed
> newFeed s i = parseAbsoluteURI s >>= (\ u -> return Feed {
>                                                 uri = u,
>                                                 interval = i,
>                                                 thread = Nothing,
>                                                 modified = Nothing,
>                                                 status = NoError,
>                                                 feedtitle = Nothing,
>                                                 feedlink = Nothing,
>                                                 feeddesc = Nothing,
>                                                 items = []
>                                               })

When we're parsing the XML, we want to pull out the channel information (title,
link, and description). All three must be present.

> pullChannel :: Maybe Content -> Feed -> Maybe Feed
> pullChannel Nothing                       _ = Nothing
> pullChannel (Just (CElem (Elem _ _ con))) f = do
>   title <- getContent con "title"
>   link <- getContent con "link"
>   desc <- getContent con "description"
>   return f { feedtitle = Just title, feedlink = Just link, feeddesc = Just desc }

If an item does not have a guid, we create one for it.

> hash :: [Maybe String] -> String
> hash = md5s . Str . foldl (++) [] . catMaybes

Given a content element, create an Item out of it. An item needs at least
either a title or a description. (Make sure we're passed in an "item" content
element; otherwise, return Nothing.)

> xmlToItem :: Content -> Maybe Item
> xmlToItem (CElem (Elem "item" _ content)) = if isNothing title && isNothing desc
>   then Nothing
>   else Just (Item { guid = mkguid,
>                     itemtitle = title,
>                     itemlink = link,
>                     itemdesc = desc,
>                     author = creator })
>   where title = getContent content "title"
>         desc = getContent content "description"
>         link = getContent content "link"
>         creator = getContent content "dc:creator"
>         mkguid = case getContent content "guid" of
>                   Just g  -> g
>                   Nothing -> hash [title, desc, link]
> xmlToItem _ = Nothing

Given an array of "item" content elements, create the items out of them.

> pullItems :: [Content] -> [Item]
> pullItems = mapMaybe xmlToItem

Given a parsed XML Document, pull out all the information about the feed that
we care about. There are slight variations between the different RSS formats
that need to be watched out for.

First try to pull out the channel information. If that fails, there's an RSS
error. If it succeeds, then the feed is fine (NoError), but we still need to
pull out the items, of which there may be none.

> pullFeed :: [Content] -> Int -> Feed -> Feed
> pullFeed c v f = case pullChannel (getChannel c) f of
>                           Nothing -> f { status = RssError }
>                           Just g  -> g { status = NoError,
>                                          items = pullItems (getItems c v) }
>   where getChannel = find (elemMatching "channel")
>         getItems c v = filter (elemMatching "item")
>                               (case v of
>                                 2 -> channelContent c
>                                 1 -> c)
>         channelContent c = case getChannel c of
>                             Nothing                     -> []
>                             Just (CElem (Elem _ _ con)) -> con

The main function for converting the XML Document to the updated feed.

> parseFeed :: Document -> Feed -> Feed
> parseFeed (Document _ _ (Elem nm _ content) _) f = case nm of
>   "rdf:RDF" -> pullFeed content 1 f
>   "rss"     -> pullFeed content 2 f
>   otherwise -> f { status = RssError }

Inside the BrowserAction monad, set the cookies, fetch the feed, and parse the
document. We need to manage the Last-Modified header ourselves.

!!! There are bugs in the Network.HTTP module:
http://bugs.debian.org/471137
http://bugs.debian.org/471140

I've "fixed" these bugs in my local version, but they're not patches that
should be applied upstream, because they're incomplete fixes just to do what I
want.

> getFeedWithCookies :: Feed -> [Cookie] -> Status -> BrowserAction Feed
> getFeedWithCookies feed cookies oldstatus = do
>   setErrHandler $ (\ s -> return ())
>   setOutHandler $ (\ s -> return ())
>   if debug then setDebugLog (Just "log") else return ()
>   setCookies cookies
>   let mod = case (modified feed) of
>               Just mo -> insertHeaderIfMissing HdrIfModifiedSince mo
>               Nothing -> id
>   let req = mod $ defaultGETRequest (uri feed)
>   (_, resp) <- request req
>   let lastmod = lookupHeader HdrLastModified (rspHeaders resp)
>   return $ case (rspCode resp) of
>     (3,0,4) -> feed { status = oldstatus }
>     (2,_,_) -> either (\ err -> feed { status = XmlError, modified = lastmod })
>                       (\ doc -> (parseFeed doc feed) { modified = lastmod })
>                       (xmlParse' (show $ uri feed) (rspBody resp))
>     _       -> feed { status = HdrError }

> cookieFile :: IO String
> cookieFile = do
>   dir <- mydir
>   return (dir ++ "/haskies")

If there are any problems reading the cookies, we'll just return an empty
array. this includes io errors, read errors, etc. We don't bother saving the
new cookies. If you want fresh cookies, bake them yourself.

> getFeed :: Feed -> Status -> IO Feed
> getFeed feed oldstatus = do
>   cookies <- catch (cookieFile >>= readFile >>= return . map read . lines) (\ e -> return [])
>   catch (browse $ getFeedWithCookies feed cookies oldstatus) (\ e -> return feed { status = NetError })

> sleep :: Int -> IO ()
> sleep = (flip replicateM_) (threadDelay (1000 * 1000 * 60))

> monitorFeed :: TVar Feed -> TVar Bool -> IO a
> monitorFeed var nr = do
>   thr <- myThreadId
>   (feed, oldstatus) <- atomically $ do
>     f <- readTVar var
>     let u = f { status = Updating, thread = Just thr }
>     writeTVar var u
>     return (u, status f)
>   atomically $ writeTVar nr True
>   upd <- getFeed feed oldstatus
>   atomically $ writeTVar var upd
>   atomically $ writeTVar nr True
>   E.catchDyn (sleep (60 * (interval feed))) (\ Restart -> return ())
>   monitorFeed var nr

vim:set sw=2 ts=2 et:
